#ifndef _GRAPHICS_H
#define _GRAPHICS_H
#include "malloc.h"
#include "wad.h"
#include "defs.h"
#define VIDEO_SEG    0xA000
#define RENDER_COLOR 130
#define FONT_HEIGHT  14

void far *screenbuf;
void far *font;

void set_mode(uint mode);
void put_pixel(uint x, uint y, uchar c);
void cls(uchar c);
void bar(int x, int y, uint w, uint h, uchar c);
void rect(int x, int y, uint w, uint h, uchar c);
void put_render_image(int x1, int y1, uint w, uint h, uint x2, uint y2, uint w_, void far *buf);
void put_image(int x1, int y1, uint w, uint h, uint x2, uint y2, uint w_, void far *buf);
void vert_line(int x, int y, uint h, uchar c);
void hor_line(int x, int y, uint w, uchar c);
//void panel(int x1, int y1, int w, int h, int c1, int c2, int c3);

void copy_to_screen();
void set_screen_buf(void far *buf);

void vert_ret();
void set_rgb(uchar n, uchar r, uchar g, uchar b);

void load_font(char *filename);
void out_char(int x, int y, uchar c, uchar color);
void out_string(int x, int y, char *s, uchar c);
void out_size_char(int x, int y, uchar c, uchar color, uchar size_x, uchar size_y);
void out_size_string(int x, int y, char *s, uchar c, uchar size_x, uchar size_y);

void set_mode(uint mode)
{
  asm mov ax, mode
  asm int 10h
}

void put_pixel(uint x, uint y, uchar c)
{
  asm
  {
    cmp x, 319
    ja @1
    cmp y, 199
    ja @1
    cmp x, 76
    jb @1
    push es
    les di, screenbuf
    mov ax, y
    mov bx, 320
    mul bx
    add ax, x
    add di, ax
    mov al, c
    stosb
    pop es
@1:
  }
}

void cls(uchar c)
{
  asm
  {
    push es
    les di, screenbuf
    mov cx, 64000
    mov al, c
    rep stosb
    pop es
  }
}

void bar(int x, int y, uint w, uint h, uchar c)
{
  if (x < 0) { w += x; x = 0; }
  if (y < 0) { h += y; y = 0; }

  if (x + w > 320) w = 320 - x;
  if (y + h > 200) h = 200 - y;

  if (w > 0 && h > 0)
  asm
  {
    cmp x, 319
    ja @2
    cmp y, 199
    ja @2
    les di, screenbuf
    mov ax, y
    mov bx, 320
    mul bx
    add ax, x
    add di, ax
    mov al, c
    mov cx, h
@1: push cx
    mov cx, w
    rep stosb
    sub di, w
    add di, 320
    pop cx
    loop @1
@2:
  }
}

void rect(int x, int y, uint w, uint h, uchar c)
{
  hor_line(x, y, w, c);
  hor_line(x, y + h, w, c);
  vert_line(x, y, h, c);
  vert_line(x + w, y, ++h, c);
}

void put_render_image(int x1, int y1, uint w, uint h, uint x2, uint y2, uint w_, void far *buf)
{
  int offs = y2 * w_ + x2;

  if ((int)(x1 + w) < 0 || (int)(y1 + h) < 0) return;

  if (y1 < 0) { offs -= w_ * y1; h += y1; y1 = 0; }
  if (x1 < 0) { offs -= x1; w += x1; x1 = 0; }

  if (x1 + w > 320) w = 320 - x1;
  if (y1 + h > 200) h = 200 - y1;

  if (w > 0 && h > 0)
  asm
  {
    cmp x1, 319
    ja @2
    cmp y1, 199
    ja @2
    push ds
    les di, screenbuf
    lds si, buf
    mov ax, y1
    mov bx, 320
    mul bx
    add ax, x1
    add si, offs
    add di, ax
    mov cx, h
@1: push cx
    mov cx, w
@3: mov al, ds:[si]
    cmp al, RENDER_COLOR
    je @4
    movsb
    jmp @5
@4: inc si
    inc di
@5: loop @3
    sub di, w
    add di, 320
    sub si, w
    add si, w_
    pop cx
    loop @1
    pop ds
@2:
  }
}

void put_image(int x1, int y1, uint w, uint h, uint x2, uint y2, uint w_, void far *buf)
{
  int offs = y2 * w_ + x2;

  if (y1 < 0) { offs -= w_ * y1; h += y1; y1 = 0; }
  if (x1 < 0) { offs -= x1; w += x1; x1 = 0; }

  if (x1 + w > 320) w = 320 - x1;
  if (y1 + h > 200) h = 200 - y1;

  if (w && h)
  asm
  {
    cmp x1, 319
    ja @2
    cmp y1, 199
    ja @2
    push ds
    les di, screenbuf
    lds si, buf
    mov ax, y1
    mov bx, 320
    mul bx
    add ax, x1
    add si, offs
    add di, ax
    mov cx, h
@1: push cx
    mov cx, w
    rep movsb
    sub di, w
    add di, 320
    sub si, w
    add si, w_
    pop cx
    loop @1
    pop ds
@2:
  }
}

void vert_line(int x, int y, uint h, uchar c)
{
  if (y < 0) { h += y; y = 0; }

  if (h)
  asm
  {
    cmp x, 319
    ja @2
    cmp y, 199
    ja @2
    les di, screenbuf
    mov ax, y
    mov bx, 320
    mul bx
    add ax, x
    add di, ax
    mov cx, h
    mov al, c
@1: stosb
    add di, 319
    loop @1
@2:
  }
}

void hor_line(int x, int y, uint w, uchar c)
{
  if (x < 0) { w += x; x = 0; }

  if (w)
  asm
  {
    cmp x, 319
    ja @2
    cmp y, 199
    ja @2
    les di, screenbuf
    mov ax, y
    mov bx, 320
    mul bx
    add ax, x
    add di, ax
    mov cx, w
    mov al, c
    rep stosb
@2:
  }
}

void copy_to_screen()
{
  asm
  {
    push es
    push ds
    lds si, screenbuf
    mov ax, VIDEO_SEG
    mov es, ax
    xor di, di
    mov cx, 32000
    rep movsw
    pop ds
    pop es
  }
}

void set_screen_buf(void far *buf)
{
  screenbuf = buf;
}

void vert_ret()
{
  asm
  {
    cli
    mov dx, 03DAh
@1: in al, dx
    test al, 8
    jz @1
@2: in al, dx
    test al, 8
    jnz @2
    sti
  }
}

void set_rgb(uchar n, uchar r, uchar g, uchar b)
{
  asm
  {
    mov dx, 3C8h
    mov al, n
    out dx, al
    inc dx
    mov al, r
    shr al, 2
    out dx, al
    mov al, g
    shr al, 2
    out dx, al
    mov al, b
    shr al, 2
    out dx, al
  }
}

void load_font(char *filename)
{
  font = farmalloc(FONT_HEIGHT * 256);
  W_Read(filename, 0, FONT_HEIGHT * 256, font);
}

void out_char(int x, int y, uchar c, uchar color)
{
  uchar b;

  for (int i = 0; i < FONT_HEIGHT; i++)
  {
    asm
    {
      push ds
      lds si, font
      xor ah, ah
      mov al, c
      mov bl, FONT_HEIGHT
      mul bl
      add ax, i
      add si, ax
      lodsb
      mov b, al
      pop ds
    }
    for (int j = 0; j < 8; j++)
      if (b & (1 << j))
        put_pixel(x + (7 - j), y + i, color);
  }
}

void out_size_char(int x, int y, uchar c, uchar color, uchar size_x, uchar size_y)
{
  uchar b;

  for (int i = 0; i < FONT_HEIGHT; i++)
  {
    asm
    {
      push ds
      lds si, font
      xor ah, ah
      mov al, c
      mov bl, FONT_HEIGHT
      mul bl
      add ax, i
      add si, ax
      lodsb
      mov b, al
      pop ds
    }
    for (int j = 0; j < 8; j++)
      if (b & (1 << j))
        bar(x + (7 - j) * size_x, y + i * size_y, size_x, size_y, color);
  }
}

void out_string(int x, int y, char *s, uchar c)
{
  int i = 0;

  while(*s)
  {
    out_char(x + i++ * 9, y, *s++, c);
  }
}

void out_size_string(int x, int y, char *s, uchar c, uchar size_x, uchar size_y)
{
  int i = 0;

  while(*s)
  {
    out_size_char(x + i++ * size_x * 9, y, *s++, c, size_x, size_y);
  }
}

/*void panel(int x1, int y1, int w, int h, int c1, int c2, int c3)
{
  bar(x1 + 2, y1 + 2, w - 4, h - 4, c3);

  hor_line(x1, y1, w - 1, c1);
  hor_line(x1 + 1, y1 + 1, w - 3, c1);
  vert_line(x1, y1 + 1, h - 1, c1);
  vert_line(x1 + 1, y1 + 2, h - 3, c1);

  hor_line(x1 + 2, y1 + h - 2, w - 4, c2);
  hor_line(x1 + 1, y1 + h - 1, w - 3, c2);
  vert_line(x1 + w - 2, y1 + 1, h - 1, c2);
  vert_line(x1 + w - 1, y1, h, c2);
} */

#endif
