#ifndef _WAD_H
#define _WAD_H
#include <stdio.h>
#include <string.h>
#include <alloc.h>
#include "error.h"

const char *magic = "WAD2";
FILE *wad;

struct WAD {
  char name[12];
  long offs;
  long size;
  WAD *next;
} *beg_table = NULL, *end_table = NULL;

void W_Init(char *wadname);
void W_Read(char *filename, long from, long size, void *buffer);
void W_Next(long size, void *buffer);
void W_Close(void);

void W_Init(char *wadname)
{
  char buf;
  long offs;
  WAD *p;

  wad = fopen(wadname, "rb");
  if (!wad) fatal("can't open wad file");

  for (int i = 0; i < 4; i++)
  {
    buf = getc(wad);
    if (buf != magic[i]) fatal("not a wad file");
  }

  fread((void *)&offs, 4, 1, wad);
  fseek(wad, offs, 0);

  while (!feof(wad))
  {
    p = (WAD *)malloc(sizeof(WAD));
    if (!p) fatal("not enough memory");
    fread((void *)p, 20, 1, wad);
    p -> next = NULL;
    if (wad)
    {
      if (!beg_table)
        beg_table = p;
      else
        end_table -> next = p;
      end_table = p;
    }
  }
}

void W_Read(char *filename, long from, long size, void *buffer)
{
  long offs = 0;
  WAD *p = beg_table;
  char name[13];

  while (p && strcmp(name, filename))
  {
    for (int i = 0; i < 12; i++)
      name[i] = p -> name[i];
    name[12] = 0;
    offs = p -> offs + from;
    p = p -> next;
  }

  if (strcmp(name, filename))
    fatal("wad subfile not found");

  fseek(wad, offs, 0);
  fread(buffer, size, 1, wad);

  if (feof(wad)) fatal("error reading wad file");
}

void W_Close(void)
{
  WAD *p = beg_table;

  while (p)
  {
    beg_table = p -> next;
    free(p);
    p = beg_table;
  }

  fclose(wad);
}

void W_Next(long size, void *buffer)
{
  fread(buffer, size, 1, wad);
  if (feof(wad)) fatal("error reading wad file");
}

#endif
